import datetime
import typing as t

import pytz

from dateutil.parser import parse as date_parse
from tzlocal import get_localzone


epoch = datetime.datetime.utcfromtimestamp(0)


def is_weekday(dt: datetime.datetime):
    return dt.weekday() < 5


def is_weekend(dt: datetime.datetime):
    return dt.weekday() >= 5


def just_now(tz: str = 'UTC'):
    if not tz:
        now = datetime.datetime.utcnow()
    else:
        now = datetime.datetime.now(tz=pytz.timezone(tz))

    return now.replace(microsecond=0)


def normalize_datetime(dt: datetime.datetime, tz: str = 'UTC'):
    local_tz = pytz.timezone(tz)
    if not dt.tzinfo:
        dt = local_tz.localize(dt)
    else:
        dt = dt.astimezone(local_tz)

    return dt.replace(microsecond=0)


def normalize_isoformat(dt: str, tz: str = 'UTC'):
    dt = date_parse(dt)
    return normalize_datetime(dt, tz)


def unix_time_millis(dt: datetime.datetime = None):
    dt = dt or datetime.datetime.utcnow()
    return int((dt - epoch).total_seconds() * 1000.0)


def unix_time_to_human(uxtime: int):
    return datetime.datetime.fromtimestamp(uxtime / 1000).strftime('%Y-%m-%d %H:%M:%S')



import collections
import datetime

import pytz
import voluptuous as v

from dateutil.relativedelta import (
    MO,
    TU,
    WE,
    TH,
    FR,
    SA,
    SU,
)


TimeUnits = collections.namedtuple(
    'TimeUnits', [
        'SECONDS',
        'MINUTES',
        'HOURS',
        'DAYS',
        'WEEKS',
        'MONTHS',
    ]
)(
    SECONDS='seconds',
    MINUTES='minutes',
    HOURS='hours',
    DAYS='days',
    WEEKS='weeks',
    MONTHS='months',
)

PeriodicalUnits = collections.namedtuple(
    'PeriodicalUnits', [
        'YEARLY',
        'MONTHLY',
        'WEEKLY',
        'DAILY',
        'HOURLY',
        'MINUTELY',
        'SECONDLY',
    ]
)(
    YEARLY='yearly',
    MONTHLY='monthly',
    WEEKLY='weekly',
    DAILY='daily',
    HOURLY='hourly',
    MINUTELY='minutely',
    SECONDLY='secondly',
)

RelativeUnits = collections.namedtuple(
    'RelativeUnits', [
        'DAY',
        'WEEKDAY',
        'WEEKEND',
        'SUNDAY',
        'MONDAY',
        'TUESDAY',
        'WEDNESDAY',
        'THURSDAY',
        'FRIDAY',
        'SATURDAY',
    ]
)(
    DAY='day',
    WEEKDAY='weekday',
    WEEKEND='weekend',
    SUNDAY='sunday',
    MONDAY='monday',
    TUESDAY='tuesday',
    WEDNESDAY='wednesday',
    THURSDAY='thursday',
    FRIDAY='friday',
    SATURDAY='saturday',
)

RelativeIndexUnits = collections.namedtuple(
    'RelativeIndexUnits', [
        'FIRST',
        'SECOND',
        'THIRD',
        'FOURTH',
        'LAST',
    ]
)(
    FIRST='first',
    SECOND='second',
    THIRD='third',
    FOURTH='fourth',
    LAST='last',
)

WeekdayUnits = collections.namedtuple(
    'WeekdayUnits', [
        'MONDAY',
        'TUESDAY',
        'WEDNESDAY',
        'THURSDAY',
        'FRIDAY',
        'SATURDAY',
        'SUNDAY',
    ]
)(
    MONDAY=MO.weekday,
    TUESDAY=TU.weekday,
    WEDNESDAY=WE.weekday,
    THURSDAY=TH.weekday,
    FRIDAY=FR.weekday,
    SATURDAY=SA.weekday,
    SUNDAY=SU.weekday,
)

MonthUnits = collections.namedtuple(
    'MonthUnits', [
        'JANUARY',
        'FEBRUARY',
        'MARCH',
        'APRIL',
        'MAY',
        'JUNE',
        'JULY',
        'AUGUST',
        'SEPTEMBER',
        'OCTOBER',
        'NOVEMBER',
        'DECEMBER',
    ]
)(
    JANUARY=1,
    FEBRUARY=2,
    MARCH=3,
    APRIL=4,
    MAY=5,
    JUNE=6,
    JULY=7,
    AUGUST=8,
    SEPTEMBER=9,
    OCTOBER=10,
    NOVEMBER=11,
    DECEMBER=12,
)

GettersSchema = v.Schema([
    {
        v.Required('getter'): object,
        'params': dict
    }
])

StartSection = {
    'on': v.Maybe(datetime.datetime),
    'relative_timeshift': {
        v.Required('delay'): v.Maybe(
            v.All(
                v.Coerce(int, msg='Invalid start timeshift delay value, an integer expected'),
                v.Range(min=1)
            )
        ),
        v.Required('time_units'): v.Maybe(
            v.All(
                str, v.In(set(TimeUnits),
                            msg='Invalid start timeshift unit value, one of %s expected' % (
                    ', '.join(['"%s"' % x for x in tuple(TimeUnits)])
                )),
            ),
        )
    }
}

StopSection = {
    v.Required('never'): v.Maybe(v.Boolean()),
    'on': v.Maybe(datetime.datetime),
    'after_num_repeats': v.Maybe(v.All(int, v.Range(min=1))),
}


RelativeScheduleSchema = v.Schema({
    'start': StartSection,
    'stop': StopSection,
    v.Required('periodical'): {
        v.Required('repeats'): v.Maybe(
            v.All(
                str,
                v.In({
                    PeriodicalUnits.MONTHLY, PeriodicalUnits.YEARLY},
                    msg='Relative datetime can be with %s rotation type only' % (
                        ' or '.join(
                            '"%s"' % x for x in (PeriodicalUnits.MONTHLY, PeriodicalUnits.YEARLY)
                        ))
                )
            ),
        ),
        v.Required('relative_day'): v.All(str, v.In(set(RelativeUnits))),
        v.Required('relative_day_index'): v.All(str, v.In(set(RelativeIndexUnits))),
        v.Optional('every', default=1): v.Maybe(v.All(int, v.Range(min=1))),
        'hour': v.Maybe(v.All(int, v.Range(min=0, max=23))),
        'minute': v.Maybe(v.All(int, v.Range(min=0, max=59))),
        'second': v.Maybe(v.All(int, v.Range(min=0, max=59))),
    },
    'timezone': v.Maybe(v.All(str, v.In(pytz.all_timezones_set)))
})

ScheduleSchema = v.Schema({
    'start': StartSection,
    'stop': StopSection,
    'periodical': {
        v.Required('repeats'): v.Maybe(v.All(str, v.In(set(PeriodicalUnits)))),
        v.Optional('every', default=1): v.Maybe(v.All(int, v.Range(min=1))),
        'month': v.Maybe(v.All(int, v.Range(min=1, max=12))),
        'day': v.Maybe(v.All(int, v.Range(min=1, max=31))),
        'weekday': v.Maybe([v.All(int, v.Range(min=0, max=6))]),
        'hour': v.Maybe(v.All(int, v.Range(min=0, max=23))),
        'minute': v.Maybe(v.All(int, v.Range(min=0, max=59))),
        'second': v.Maybe(v.All(int, v.Range(min=0, max=59))),
        'relative_day': v.Maybe(v.All(str, v.In(set(RelativeUnits)))),
        'relative_day_index': v.Maybe(v.All(str, v.In(set(RelativeIndexUnits)))),
    },
    'timezone': v.Maybe(v.All(str, v.In(pytz.all_timezones_set)))
})



import math
import calendar
import datetime
import typing as t

from toolz.dicttoolz import get_in, assoc_in
from dateutil.relativedelta import (
    relativedelta,
    MO,
    TU,
    WE,
    TH,
    FR,
    SA,
    SU,
)
from dateutil.rrule import (
    rrule,
    YEARLY,
    MONTHLY,
    WEEKLY,
    DAILY,
    HOURLY,
    MINUTELY,
    SECONDLY,
)

from .structs import (
    TimeUnits,
    PeriodicalUnits,
    RelativeUnits,
    RelativeIndexUnits,
    ScheduleSchema,
    RelativeScheduleSchema,
    GettersSchema,
)
from .utils import (
    just_now,
    is_weekday,
    is_weekend,
    normalize_datetime,
)


TIMESHIFT_MAP = {
    TimeUnits.DAYS: lambda val: datetime.timedelta(days=val),
    TimeUnits.SECONDS: lambda val: datetime.timedelta(seconds=val),
    TimeUnits.MINUTES: lambda val: datetime.timedelta(minutes=val),
    TimeUnits.HOURS: lambda val: datetime.timedelta(hours=val),
    TimeUnits.WEEKS: lambda val: datetime.timedelta(weeks=val),
    TimeUnits.MONTHS: lambda val: datetime.timedelta(weeks=val * 4),
}

PERIODICAL_MAP = {
    PeriodicalUnits.YEARLY: YEARLY,
    PeriodicalUnits.MONTHLY: MONTHLY,
    PeriodicalUnits.WEEKLY: WEEKLY,
    PeriodicalUnits.DAILY: DAILY,
    PeriodicalUnits.HOURLY: HOURLY,
    PeriodicalUnits.MINUTELY: MINUTELY,
    PeriodicalUnits.SECONDLY: SECONDLY,
}

PERIODICAL_ATTRS_MAP = {
    'every': 'interval',
    'month': 'bymonth',
    'day': 'bymonthday',
    'weekday': 'byweekday',
    'hour': 'byhour',
    'minute': 'byminute',
    'second': 'bysecond',
}

SENSITIVE_ATTRS_MAP = {
    PeriodicalUnits.YEARLY: {'every', 'weekday', 'month', 'day', 'hour', 'minute', 'second'},
    PeriodicalUnits.MONTHLY: {'every', 'weekday', 'day', 'hour', 'minute', 'second'},
    PeriodicalUnits.WEEKLY: {'every', 'weekday', 'hour', 'minute', 'second'},
    PeriodicalUnits.DAILY: {'every', 'hour', 'minute', 'second'},
    PeriodicalUnits.HOURLY: {'every', 'minute', 'second'},
    PeriodicalUnits.MINUTELY: {'every', 'second'},
    PeriodicalUnits.SECONDLY: {'every'},
}

RELATIVE_DAY_MAP = {
    RelativeUnits.SUNDAY: SU,
    RelativeUnits.MONDAY: MO,
    RelativeUnits.TUESDAY: TU,
    RelativeUnits.WEDNESDAY: WE,
    RelativeUnits.THURSDAY: TH,
    RelativeUnits.FRIDAY: FR,
    RelativeUnits.SATURDAY: SA,
}


def validated_schedule(
        schedule: dict,
        getters: t.Optional[t.List[dict]] = None,
) -> dict:
    """Validates ``schedule`` with schemas and returns modified
    schedule dict if some getters provided.

    Getters are special functions that can be attached to get some
    schedule value from external source by key-path::

        getters = [
            {
                'getter': payload_getter,
                'params': {
                    'path': ['start', 'relative_timeshift', 'delay'],
                    'source': payload_source,
                }
            }
        ]

    In this case, :func:`payload_getter` will receive value by ``path``
    from the ``payload_source``. And ``schedule`` will be modified with
    the result from ``payload_getter`` by the same path.
    """
    if getters:
        getters = GettersSchema(getters)
        for getter in getters:
            modifier = getter['getter']
            params = getter.get('params', {})
            path_to_value = params.get('path')
            if path_to_value:
                path_value = get_in(path_to_value, schedule)
                if path_value is not None:
                    val = modifier(path_value, **params)
                    schedule = assoc_in(schedule, path_to_value, val)

    relative_params = (
        get_in(['periodical', 'relative_day'], schedule) and
        get_in(['periodical', 'relative_day_index'], schedule)
    )
    if relative_params:
        schedule = RelativeScheduleSchema(schedule)
    else:
        schedule = ScheduleSchema(schedule)

    return schedule


def schedule_parser(
    schedule: dict,
    now_dt: t.Optional[datetime.datetime] = None,
    getters: t.Optional[t.List[dict]] = None,
) -> t.Generator[datetime.datetime, None, None]:
    """Generates datetime objects by provided schedule structure.

    Simple periodical example::

        schedule = {
            'start': {
                'on': just_now(),
            },
            'periodical': {
                'repeats': 'weekly',
                'every': 1,
            },
            'stop': {
                'never': False,
                'after_num_repeats': 2
            }
        }
        schedule_gen = schedule_parser(schedule)


    Getters case::

        schedule_gen = schedule_parser(
            schedule,
            getters=[
                {
                    'getter': payload_getter,
                    'params': {
                        'path': ['start', 'relative_timeshift', 'delay'],
                        'source': payload_source,
                    }
                }
            ]
        )

    Pay special attention to the timezone hell. By default, parser uses UTC
    everywhere. You can set some local timezone with special ``timezone``
    field::

        start_on_jakarta = normalize_isoformat(
            '2019-03-20T22:00:00',
            tz='Asia/Jakarta',
        )
        stop_on_jakarta = normalize_isoformat(
            '2019-03-22T23:16:00',
            tz='Asia/Jakarta',
        )
        schedule = {
            'start': {
                'on': start_on_jakarta,
            },
            'periodical': {
                'repeats': 'daily'
                'hour': 23,
                'minute': 15,
                'every': 1,
            },
            'stop': {
                'never': False,
                'on': stop_on_jakarta,
            },
            'timezone': 'Asia/Jakarta',
        }
        schedule_gen = schedule_parser(schedule)

    If start and stop datetime objects signed with some tz (not naive),
    that will be changed according to the timezone from this field.

    If datetime objects are naive, they will be signed with timezone.
    If there is no ``timezone`` — UTC is the default one.
    """
    schedule = validated_schedule(schedule, getters=getters)

    explicit_tz = schedule.get('timezone', 'UTC')
    if now_dt:
        now = normalize_datetime(now_dt, explicit_tz)
    else:
        now = just_now(explicit_tz)

    schedule_date = get_in(['start', 'on'], schedule)
    if schedule_date:
        schedule_date = normalize_datetime(schedule_date, explicit_tz)
    else:
        schedule_date = now

    timeshift_delay = get_in(['start', 'relative_timeshift', 'delay'], schedule)
    timeshift_type = get_in(['start', 'relative_timeshift', 'time_units'], schedule)
    if timeshift_delay and timeshift_type:
        timeshift_modifier = TIMESHIFT_MAP[timeshift_type]
        schedule_date = schedule_date + timeshift_modifier(timeshift_delay)

    periodical_type = get_in(['periodical', 'repeats'], schedule)
    if not periodical_type:
        yield schedule_date

    rrule_params = {'dtstart': schedule_date}

    stop_dt = get_in(['stop', 'on'], schedule)
    if stop_dt:
        stop_dt = normalize_datetime(stop_dt, explicit_tz)
        rrule_params['until'] = stop_dt

    num_repeats = get_in(['stop', 'after_num_repeats'], schedule)
    is_infinite = get_in(['stop', 'never'], schedule)
    if num_repeats and not is_infinite:
        rrule_params['count'] = num_repeats

    if periodical_type:
        rrule_params['freq'] = PERIODICAL_MAP[periodical_type]
        context_params = SENSITIVE_ATTRS_MAP[periodical_type]
        for param in context_params:
            mapped_param = PERIODICAL_ATTRS_MAP[param]
            related_param = get_in(['periodical', param], schedule)
            rrule_params[mapped_param] = related_param

        relative_params = (
            get_in(['periodical', 'relative_day'], schedule) and
            get_in(['periodical', 'relative_day_index'], schedule)
        )

        schedule_gen = rrule(**rrule_params)
        for dt in schedule_gen:
            if dt <= now and not relative_params:
                continue

            if not relative_params:
                yield dt

            if relative_params:
                # basic case for the next planned time shift
                next_dt = relative_datetime_schedule(dt, schedule)
                if next_dt <= schedule_date:
                    continue

                yield next_dt


def relative_datetime_schedule(schedule_date, schedule_struct):
    relative_day = get_in(['periodical', 'relative_day'], schedule_struct)
    relative_day_index = get_in(['periodical', 'relative_day_index'], schedule_struct)
    repeats_type = get_in(['periodical', 'repeats'], schedule_struct)

    if repeats_type == PeriodicalUnits.MONTHLY:
        this_month_range = calendar.monthrange(schedule_date.year, schedule_date.month)
        first_day = schedule_date + relativedelta(day=1)
        _, last_day = this_month_range
        last_day = schedule_date + relativedelta(day=last_day)
    elif repeats_type == PeriodicalUnits.YEARLY:
        last_month_range = calendar.monthrange(schedule_date.year, 12)
        first_day = schedule_date + relativedelta(day=1, month=1)
        _, last_day = last_month_range
        last_day = schedule_date + relativedelta(day=last_day, month=12)

    if relative_day == RelativeUnits.DAY:
        if relative_day_index == RelativeIndexUnits.FIRST:
            schedule_date = first_day
        elif relative_day_index == RelativeIndexUnits.SECOND:
            schedule_date = first_day + relativedelta(day=2)
        elif relative_day_index == RelativeIndexUnits.THIRD:
            schedule_date = first_day + relativedelta(day=3)
        elif relative_day_index == RelativeIndexUnits.FOURTH:
            schedule_date = first_day + relativedelta(day=4)
        elif relative_day_index == RelativeIndexUnits.LAST:
            schedule_date = last_day

    elif relative_day in RELATIVE_DAY_MAP:
        weekday_modifier = RELATIVE_DAY_MAP[relative_day]
        if relative_day_index in {
            RelativeIndexUnits.FIRST,
            RelativeIndexUnits.SECOND,
            RelativeIndexUnits.THIRD,
            RelativeIndexUnits.FOURTH,
        }:
            modifier_shift_map = {
                RelativeIndexUnits.FIRST: +1,
                RelativeIndexUnits.SECOND: +2,
                RelativeIndexUnits.THIRD: +3,
                RelativeIndexUnits.FOURTH: +4,
            }
            base_date = first_day
        elif relative_day_index in {RelativeIndexUnits.LAST}:
            modifier_shift_map = {
                RelativeIndexUnits.LAST: -1,
            }
            base_date = last_day

        schedule_date = base_date + relativedelta(
            weekday=weekday_modifier(modifier_shift_map[relative_day_index])
        )

    elif relative_day in {RelativeUnits.WEEKDAY, RelativeUnits.WEEKEND}:
        if relative_day == RelativeUnits.WEEKDAY:
            is_proper_daytype = is_weekday
        elif relative_day == RelativeUnits.WEEKEND:
            is_proper_daytype = is_weekend

        if relative_day_index in {
            RelativeIndexUnits.FIRST,
            RelativeIndexUnits.SECOND,
            RelativeIndexUnits.THIRD,
            RelativeIndexUnits.FOURTH,
        }:
            days_range = range(1, last_day.day + 1)
            index_unit_slice_map = {
                RelativeIndexUnits.FIRST: 1,
                RelativeIndexUnits.SECOND: 2,
                RelativeIndexUnits.THIRD: 3,
                RelativeIndexUnits.FOURTH: 4,
            }
            base_date = first_day
        elif relative_day_index in {RelativeIndexUnits.LAST}:
            days_range = range(last_day.day, 0, -1)
            index_unit_slice_map = {
                RelativeIndexUnits.LAST: 1,
            }
            base_date = last_day

        weekday_pos = 0
        for day_num in days_range:
            maybe_this_dt = base_date + relativedelta(day=day_num)
            if is_proper_daytype(maybe_this_dt):
                weekday_pos += 1
            if index_unit_slice_map[relative_day_index] == weekday_pos:
                schedule_date = maybe_this_dt
                break

    return schedule_date


def schedule_delta(
    schedule: dict,
    now_dt: t.Optional[datetime.datetime] = None,
    getters: t.Optional[t.List[dict]] = None,
) -> t.Tuple[int, datetime.datetime]:
    schedule_gen = schedule_parser(schedule, now_dt=now_dt, getters=getters)
    explicit_tz = schedule.get('timezone', 'UTC')
    if now_dt:
        now = normalize_datetime(now_dt, explicit_tz)
    else:
        now = just_now(explicit_tz)

    try:
        schedule_dt_base = next(schedule_gen)
    except StopIteration:
        return 0, now

    schedule_seconds = (schedule_dt_base - now).total_seconds()

    if schedule_seconds <= 0:
        try:
            schedule_dt_base = next(schedule_gen)
        except StopIteration:
            return schedule_seconds, now

    schedule_seconds = (schedule_dt_base - now).total_seconds()

    return math.ceil(schedule_seconds), schedule_dt_base



import datetime

import pytest
import pytz
from voluptuous import Invalid as SchemaInvalid

from dateutil.relativedelta import relativedelta
from toolz.dicttoolz import dissoc

from krolib.utils import (
    just_now,
    normalize_datetime,
    normalize_isoformat,
)
from krolib.parser import (
    schedule_parser,
    schedule_delta,
)
from krolib.structs import (
    PeriodicalUnits,
    TimeUnits,
    RelativeUnits,
    RelativeIndexUnits,
    WeekdayUnits,
    MonthUnits,
)


@pytest.mark.unit
class TestStartTimeshiftDelay:

    def test_seconds_delay(self):
        now = just_now()
        schedule = {
            'start': {
                'relative_timeshift': {
                    'delay': 1,
                    'time_units': TimeUnits.SECONDS,
                }
            },
            'stop': {
                'never': True,
            }
        }
        schedule_gen = schedule_parser(schedule)
        result = next(schedule_gen)
        assert (result - now).total_seconds() == 1

    def test_minutes_delay(self):
        now = just_now()
        schedule = {
            'start': {
                'relative_timeshift': {
                    'delay': 1,
                    'time_units': TimeUnits.MINUTES,
                }
            },
        }
        schedule_gen = schedule_parser(schedule)
        result = next(schedule_gen)
        assert (result - now).total_seconds() == 60

    def test_hours_delay(self):
        now = just_now()
        schedule = {
            'start': {
                'relative_timeshift': {
                    'delay': 1,
                    'time_units': TimeUnits.HOURS,
                }
            },
        }
        schedule_gen = schedule_parser(schedule)
        result = next(schedule_gen)
        assert (result - now).total_seconds() == 3600

    def test_days_delay(self):
        now = just_now()
        schedule = {
            'start': {
                'relative_timeshift': {
                    'delay': 1,
                    'time_units': TimeUnits.DAYS,
                }
            },
        }
        schedule_gen = schedule_parser(schedule)
        result = next(schedule_gen)
        assert (result - now).total_seconds() == 86400

    def test_weeks_delay(self):
        now = just_now()
        schedule = {
            'start': {
                'relative_timeshift': {
                    'delay': 1,
                    'time_units': TimeUnits.WEEKS,
                }
            },
        }
        schedule_gen = schedule_parser(schedule)
        result = next(schedule_gen)
        assert (result - now).total_seconds() == 604800

    def test_months_delay(self):
        now = just_now()
        schedule = {
            'start': {
                'relative_timeshift': {
                    'delay': 1,
                    'time_units': TimeUnits.MONTHS,
                }
            },
        }
        schedule_gen = schedule_parser(schedule)
        result = next(schedule_gen)
        assert (result - now).total_seconds() == 2419200

    def test_invalid_delay(self):
        schedule = {
            'start': {
                'relative_timeshift': {
                    'delay': 1,
                    'time_units': 'invalid'
                }
            },
        }
        with pytest.raises(SchemaInvalid):
            schedule_gen = schedule_parser(schedule)
            next(schedule_gen)


@pytest.mark.unit
class TestStartDateDelay:

    def test_absolute_day_delay(self):
        now = just_now()
        start_on = now + datetime.timedelta(days=1)
        schedule = {
            'start': {
                'on': start_on,
            },
        }
        schedule_gen = schedule_parser(schedule)
        result = next(schedule_gen)
        assert (result - now).total_seconds() == 86400

    def test_absolute_day_with_shift_delay(self):
        now = just_now()
        start_on = now + datetime.timedelta(days=1)
        schedule = {
            'start': {
                'on': start_on,
                'relative_timeshift': {
                    'delay': 1,
                    'time_units': TimeUnits.SECONDS,
                }
            },
        }
        schedule_gen = schedule_parser(schedule)
        result = next(schedule_gen)
        assert (result - now).total_seconds() == 86401

    def test_absolute_past_day_delay(self):
        now = just_now()
        start_on = now + datetime.timedelta(days=-1)
        schedule = {
            'start': {
                'on': start_on,
            },
        }
        schedule_gen = schedule_parser(schedule)
        result = next(schedule_gen)
        assert result == start_on


@pytest.mark.unit
class TestPeriodicalSchedule:

    def test_secondly_periodical(self):
        now = just_now()
        schedule = {
            'periodical': {
                'repeats': PeriodicalUnits.SECONDLY,
                'every': 1,
            },
        }
        schedule_gen = schedule_parser(schedule)
        results = [next(schedule_gen) for _ in range(3)]
        deltas = [(x - now).total_seconds() for x in results]
        assert deltas == [1, 2, 3]

    def test_minutely_periodical(self):
        now = just_now()
        schedule = {
            'periodical': {
                'repeats': PeriodicalUnits.MINUTELY,
                'every': 1,
            },
        }
        schedule_gen = schedule_parser(schedule)
        results = [next(schedule_gen) for _ in range(3)]
        deltas = [(x - now).total_seconds() for x in results]
        assert deltas == [60, 120, 180]

    def test_hourly_periodical(self):
        now = just_now()
        schedule = {
            'periodical': {
                'repeats': PeriodicalUnits.HOURLY,
                'every': 1,
            },
        }
        schedule_gen = schedule_parser(schedule)
        results = [next(schedule_gen) for _ in range(3)]
        deltas = [(x - now).total_seconds() for x in results]
        assert deltas == [3600, 7200, 10800]

    def test_daily_periodical(self):
        now = just_now()
        schedule = {
            'periodical': {
                'repeats': PeriodicalUnits.DAILY,
                'every': 1,
            },
        }
        schedule_gen = schedule_parser(schedule)
        results = [next(schedule_gen) for _ in range(3)]
        deltas = [(x - now).total_seconds() / 24 for x in results]
        assert deltas == [3600, 7200, 10800]

    def test_weekly_periodical(self):
        now = just_now()
        schedule = {
            'periodical': {
                'repeats': PeriodicalUnits.WEEKLY,
                'every': 1,
            },
        }
        schedule_gen = schedule_parser(schedule)
        results = [next(schedule_gen) for _ in range(3)]
        deltas = [(x - now).total_seconds() / 7 / 24 for x in results]
        assert deltas == [3600, 7200, 10800]

    def test_monthly_periodical(self):
        now = datetime.datetime(2018, 5, 20, 12, 30, 0, tzinfo=pytz.UTC)
        next_month = now + relativedelta(months=+1)
        schedule = {
            'periodical': {
                'repeats': PeriodicalUnits.MONTHLY,
                'every': 1,
                'day': 20,
                'hour': 12,
                'minute': 30
            },
        }
        schedule_gen = schedule_parser(schedule, now_dt=now)
        result = next(schedule_gen)
        assert next_month == result.replace(second=0)

    def test_yearly_periodical(self):
        now = just_now()
        next_year = now + relativedelta(years=+1)
        schedule = {
            'periodical': {
                'repeats': PeriodicalUnits.YEARLY,
                'every': 1,
            },
        }
        schedule_gen = schedule_parser(schedule)
        result = next(schedule_gen)
        assert next_year == result

    def test_yearly_concrete_periodical(self):
        now = datetime.datetime(2018, 5, 28, tzinfo=pytz.UTC)
        schedule = {
            'periodical': {
                'repeats': PeriodicalUnits.YEARLY,
                'every': 1,
                'month': MonthUnits.JANUARY,
                'day': 1,
                'hour': 12,
                'minute': 30,
            },
        }
        schedule_gen = schedule_parser(schedule, now_dt=now)
        results = [next(schedule_gen) for _ in range(3)]
        assert results == [
            datetime.datetime(2019, 1, 1, 12, 30, tzinfo=pytz.UTC),
            datetime.datetime(2020, 1, 1, 12, 30, tzinfo=pytz.UTC),
            datetime.datetime(2021, 1, 1, 12, 30, tzinfo=pytz.UTC),
        ]

    def test_monthly_concrete_periodical(self):
        now = datetime.datetime(2018, 1, 1, tzinfo=pytz.UTC)
        schedule = {
            'periodical': {
                'repeats': PeriodicalUnits.MONTHLY,
                'every': 1,
                'weekday': [WeekdayUnits.FRIDAY],
                'day': 13,
            },
        }
        schedule_gen = schedule_parser(schedule, now_dt=now)
        results = [next(schedule_gen) for _ in range(3)]
        assert results == [
            datetime.datetime(2018, 4, 13, 0, 0, tzinfo=pytz.UTC),
            datetime.datetime(2018, 7, 13, 0, 0, tzinfo=pytz.UTC),
            datetime.datetime(2019, 9, 13, 0, 0, tzinfo=pytz.UTC),
        ]

    def test_weekly_concrete_periodical(self):
        now = datetime.datetime(2018, 5, 1, tzinfo=pytz.UTC)
        schedule = {
            'periodical': {
                'repeats': PeriodicalUnits.WEEKLY,
                'every': 1,
                'weekday': [WeekdayUnits.MONDAY, WeekdayUnits.TUESDAY],
                'hour': 23,
                'minute': 15,
                'second': 30,
            }
        }
        schedule_gen = schedule_parser(schedule, now_dt=now)
        results = [next(schedule_gen) for _ in range(3)]
        assert results == [
            datetime.datetime(2018, 5, 1, 23, 15, 30, tzinfo=pytz.UTC),
            datetime.datetime(2018, 5, 7, 23, 15, 30, tzinfo=pytz.UTC),
            datetime.datetime(2018, 5, 8, 23, 15, 30, tzinfo=pytz.UTC),
        ]

    def test_daily_concrete_periodical(self):
        now = datetime.datetime(2018, 5, 1, 1, 20, 45, tzinfo=pytz.UTC)
        schedule = {
            'periodical': {
                'repeats': PeriodicalUnits.DAILY,
                'every': 2,
                'hour': 1,
                'minute': 30,
                'second': 45,
            }
        }
        schedule_gen = schedule_parser(schedule, now_dt=now)
        results = [next(schedule_gen) for _ in range(4)]
        assert results == [
            datetime.datetime(2018, 5, 1, 1, 30, 45, tzinfo=pytz.UTC),
            datetime.datetime(2018, 5, 3, 1, 30, 45, tzinfo=pytz.UTC),
            datetime.datetime(2018, 5, 5, 1, 30, 45, tzinfo=pytz.UTC),
            datetime.datetime(2018, 5, 7, 1, 30, 45, tzinfo=pytz.UTC),
        ]


@pytest.mark.unit
class TestRelativeSchedule:

    def test_relative_first_monday_monthly(self):
        now = datetime.datetime(year=2018, month=5, day=1)
        schedule = {
            'periodical': {
                'repeats': PeriodicalUnits.MONTHLY,
                'relative_day': RelativeUnits.MONDAY,
                'relative_day_index': RelativeIndexUnits.FIRST,
            },
        }
        schedule_gen = schedule_parser(schedule, now_dt=now)
        results = [next(schedule_gen) for _ in range(3)]
        assert results == [
            datetime.datetime(2018, 5, 7, 0, 0, tzinfo=pytz.UTC),
            datetime.datetime(2018, 6, 4, 0, 0, tzinfo=pytz.UTC),
            datetime.datetime(2018, 7, 2, 0, 0, tzinfo=pytz.UTC),
        ]

    def test_relative_first_monday_yearly(self):
        now = datetime.datetime(year=2018, month=5, day=1)
        schedule = {
            'periodical': {
                'repeats': PeriodicalUnits.YEARLY,
                'relative_day': RelativeUnits.MONDAY,
                'relative_day_index': RelativeIndexUnits.FIRST,
            }
        }
        schedule_gen = schedule_parser(schedule, now_dt=now)
        results = [next(schedule_gen) for _ in range(3)]
        assert results == [
            datetime.datetime(2019, 1, 7, 0, 0, tzinfo=pytz.UTC),
            datetime.datetime(2020, 1, 6, 0, 0, tzinfo=pytz.UTC),
            datetime.datetime(2021, 1, 4, 0, 0, tzinfo=pytz.UTC),
        ]

    def test_relative_last_monday_monthly(self):
        now = datetime.datetime(year=2018, month=5, day=1)
        schedule = {
            'periodical': {
                'repeats': PeriodicalUnits.MONTHLY,
                'relative_day': RelativeUnits.MONDAY,
                'relative_day_index': RelativeIndexUnits.LAST,
            },
        }
        schedule_gen = schedule_parser(schedule, now_dt=now)
        results = [next(schedule_gen) for _ in range(3)]
        assert results == [
            datetime.datetime(2018, 5, 28, 0, 0, tzinfo=pytz.UTC),
            datetime.datetime(2018, 6, 25, 0, 0, tzinfo=pytz.UTC),
            datetime.datetime(2018, 7, 30, 0, 0, tzinfo=pytz.UTC),
        ]

    def test_relative_first_weekday_monthly(self):
        now = datetime.datetime(year=2018, month=5, day=1)
        schedule = {
            'periodical': {
                'repeats': PeriodicalUnits.MONTHLY,
                'hour': 1,
                'relative_day': RelativeUnits.WEEKDAY,
                'relative_day_index': RelativeIndexUnits.FIRST,
            }
        }
        schedule_gen = schedule_parser(schedule, now_dt=now)
        results = [next(schedule_gen) for _ in range(3)]
        assert results == [
            datetime.datetime(2018, 5, 1, 1, 0, tzinfo=pytz.UTC),
            datetime.datetime(2018, 6, 1, 1, 0, tzinfo=pytz.UTC),
            datetime.datetime(2018, 7, 2, 1, 0, tzinfo=pytz.UTC),
        ]

    def test_relative_first_weekend_monthly(self):
        now = datetime.datetime(year=2018, month=1, day=1)
        schedule = {
            'periodical': {
                'repeats': PeriodicalUnits.MONTHLY,
                'relative_day': RelativeUnits.WEEKEND,
                'relative_day_index': RelativeIndexUnits.FIRST,
            }
        }
        schedule_gen = schedule_parser(schedule, now_dt=now)
        results = [next(schedule_gen) for _ in range(3)]
        assert results == [
            datetime.datetime(2018, 1, 6, 0, 0, tzinfo=pytz.UTC),
            datetime.datetime(2018, 2, 3, 0, 0, tzinfo=pytz.UTC),
            datetime.datetime(2018, 3, 3, 0, 0, tzinfo=pytz.UTC),
        ]

    def test_relative_last_day_monthly(self):
        now = datetime.datetime(year=2018, month=1, day=1)
        schedule = {
            'periodical': {
                'repeats': PeriodicalUnits.MONTHLY,
                'relative_day': RelativeUnits.DAY,
                'relative_day_index': RelativeIndexUnits.LAST,
            }
        }
        schedule_gen = schedule_parser(schedule, now_dt=now)
        results = [next(schedule_gen) for _ in range(3)]
        assert results == [
            datetime.datetime(2018, 1, 31, 0, 0, tzinfo=pytz.UTC),
            datetime.datetime(2018, 2, 28, 0, 0, tzinfo=pytz.UTC),
            datetime.datetime(2018, 3, 31, 0, 0, tzinfo=pytz.UTC),
        ]

    def test_yearly_with_stop_periodical(self):
        now = just_now()
        next_year = now + relativedelta(years=+1)
        schedule = {
            'periodical': {
                'repeats': PeriodicalUnits.YEARLY,
                'every': 1,
            },
            'stop': {
                'on': now + datetime.timedelta(days=400),
                'never': False
            }
        }
        schedule_gen = schedule_parser(schedule)
        result = next(schedule_gen)
        assert next_year == result

        with pytest.raises(StopIteration):
            next(schedule_gen)

    def test_weekly_with_repeats_periodical(self):
        now = just_now()
        start_on = now + datetime.timedelta(days=1)
        schedule = {
            'start': {
                'on': start_on,
            },
            'periodical': {
                'repeats': PeriodicalUnits.WEEKLY,
                'every': 1,
            },
            'stop': {
                'never': False,
                'after_num_repeats': 2
            }
        }
        schedule_gen = schedule_parser(schedule)
        one_dt, two_dt = list(schedule_gen)
        assert one_dt == start_on
        assert (two_dt - one_dt).days == 7


@pytest.mark.unit
class TestTimeUtils:

    def test_timezone_eq(self):
        aware_local = just_now(tz='Europe/Kiev')
        unaware_utc = datetime.datetime.utcnow()
        aware_utc = just_now()

        n_aware_local = normalize_datetime(aware_local)
        n_unaware_utc = normalize_datetime(unaware_utc)
        n_aware_utc = normalize_datetime(aware_utc)

        assert n_aware_local == n_unaware_utc == n_aware_utc

    def test_normalizers_eq(self):
        dt_obj = datetime.datetime(2018, 5, 26, 11, 45)
        dt_str = '2018-05-26T11:45'
        assert normalize_datetime(dt_obj) == normalize_isoformat(dt_str)


@pytest.mark.unit
class TestMixedSchedule:

    def test_kiev_timezone_apply(self):
        now = just_now()
        start_on = now + datetime.timedelta(days=1)
        schedule = {
            'start': {
                'on': start_on,
            },
            'timezone': 'Europe/Kiev',
        }
        schedule_gen = schedule_parser(schedule)
        result = next(schedule_gen)
        assert (result - now).total_seconds() == 86400

    def test_jakarta_timezone_apply(self):
        now_in_kiev = normalize_isoformat(
            '2019-03-05T17:00:00',
            tz='Europe/Kiev',
        )
        start_on_jakarta = normalize_isoformat(
            '2019-03-05T22:00:00',
            tz='Asia/Jakarta',
        )

        assert not bool(start_on_jakarta - now_in_kiev)  # same time

        start_on_jakarta = normalize_isoformat(
            '2019-03-05T15:05:00',
        )

        schedule = {
            'start': {
                'on': start_on_jakarta,
                'relative_timeshift': {
                    'delay': None,
                    'time_units': None,
                }
            },
            'stop': {
                'on': None,
                'never': None,
                'after_num_repeats': None,
            },
            'timezone': "Asia/Jakarta"
        }
        schedule_gen = schedule_parser(schedule)
        result = next(schedule_gen)

        assert (result - now_in_kiev).seconds == 300

    def test_different_timezones(self):
        now = just_now(tz=None)
        start_on = now + datetime.timedelta(days=1)
        kyiv_schedule = {
            'start': {
                'on': start_on,
            },
            'timezone': 'Europe/Kiev',
        }
        kyiv_schedule_gen = schedule_parser(kyiv_schedule)
        kyiv_result = next(kyiv_schedule_gen)

        warsaw_schedule = {
            'start': {
                'on': start_on,
            },
            'timezone': 'Europe/Warsaw',
        }
        warsaw_schedule_gen = schedule_parser(warsaw_schedule)
        warsaw_result = next(warsaw_schedule_gen)

        assert (warsaw_result - kyiv_result).total_seconds() == 3600

    def test_start_periodical_same_time(self):
        now = just_now(tz='Europe/Kiev')
        start_on = now + datetime.timedelta(days=1)
        start_on = start_on.replace(hour=11, minute=45)

        schedule = {
            'timezone': 'Europe/Kiev',
            'start': {
                'on': start_on,
            },
            'periodical': {
                'repeats': 'daily',
                'every': 1,
                'hour': 11,
                'minute': 45,
            },
            'stop': {
                'never': True,
            }
        }
        schedule_gen = schedule_parser(schedule)
        results = [next(schedule_gen) for _ in range(3)]
        assert results == [
            start_on,
            start_on + datetime.timedelta(days=1),
            start_on + datetime.timedelta(days=2),
        ]

    def test_periodical_time_for_custom_tz(self):
        start_on_jakarta = normalize_isoformat(
            '2019-03-20T22:00:00',
            tz='Asia/Jakarta',
        )
        stop_on_jakarta = normalize_isoformat(
            '2019-03-22T23:16:00',
            tz='Asia/Jakarta',
        )

        schedule = {
            'start': {
                'on': start_on_jakarta,
            },
            'periodical': {
                'repeats': PeriodicalUnits.DAILY,
                'hour': 23,
                'minute': 15,
                'every': 1,
            },
            'stop': {
                'never': False,
                'on': stop_on_jakarta,
            },
            'timezone': 'Asia/Jakarta',
        }
        schedule_gen = schedule_parser(schedule, now_dt=start_on_jakarta)
        results = list(schedule_gen)
        assert [s.tzname() for s in results] == ['WIB', 'WIB', 'WIB']

        one, two, three = results
        assert (one - start_on_jakarta).total_seconds() == 4500  # 75 minutes, start
        assert (two - one).total_seconds() == 86400  # 1 day
        assert (three - two).total_seconds() == 86400  # 1 day
        assert three == normalize_isoformat('2019-03-22T23:15:00', tz='Asia/Jakarta')


@pytest.mark.unit
class TestScheduleDelta:

    def test_hours_delay_delta(self):
        schedule = {
            'start': {
                'relative_timeshift': {
                    'delay': 1,
                    'time_units': TimeUnits.HOURS,
                }
            },
        }
        delta, _ = schedule_delta(schedule)
        assert delta == 3600

    def test_delta_with_delay(self):
        schedule = {
            'start': {
                'relative_timeshift': {
                    'delay': '1',
                    'time_units': TimeUnits.DAYS,
                }
            },
            'periodical': {
                'repeats': PeriodicalUnits.HOURLY,
                'every': 1,
            },
            'timezone': 'Europe/Kiev',
        }
        delta, _ = schedule_delta(schedule)
        abs_delta, _ = schedule_delta(dissoc(schedule, 'start'))
        assert delta == 86400
        assert abs_delta == 3600

    def test_delta_with_stop(self):
        now = just_now()
        schedule = {
            'periodical': {
                'repeats': PeriodicalUnits.YEARLY,
                'every': 1,
            },
            'stop': {
                'on': now + datetime.timedelta(days=1),
                'never': False
            }
        }
        delta, _ = schedule_delta(schedule)
        assert delta == 0



import io
import os
import pathlib
import sys
from shutil import rmtree
from setuptools import setup, find_packages, Command


HERE_PATH = pathlib.Path(__file__).parent


NAME = 'krolib'
DESCRIPTION = 'DSL to handle complex schedules'
URL = 'https://github.com/nextiva/krolib'
EMAIL = 'roman.zayev@nextiva.com'
AUTHOR = 'Nextiva'
REQUIRES_PYTHON = '>=3.6.0'
VERSION = '1.2.0'


try:
    with io.open((HERE_PATH / 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(HERE_PATH / 'dist', ignore_errors=True)
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(VERSION))
        os.system('git push --tags')

        sys.exit()

setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=EMAIL,
    packages=find_packages(),
    license='Apache 2.0',
    url=URL,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        'python-dateutil==2.8.0',
        'pytz==2019.1',
        'toolz>=0.8.2',
        'tzlocal>=1.3',
        'voluptuous==0.11.5',
    ],
    tests_require=[
        'pytest==5.0.1',
        'pytest-asyncio==0.10.0',
        'pytest-cov==2.7.1',
        'pytest-flakes==4.0.0',
        'pytest-pycodestyle==1.4.0',
    ],
    setup_requires=['pytest-runner'],
    python_requires=REQUIRES_PYTHON,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    cmdclass={
        'upload': UploadCommand,
    },
)



import asyncio
import pytest
from voluptuous import Invalid as SchemaInvalid

from krolib.asyncio import scheduler
from krolib.structs import TimeUnits, PeriodicalUnits


pytestmark = pytest.mark.asyncio


async def test_scheduler(event_loop):

    @scheduler({
        'start': {
            'relative_timeshift': {
                'delay': 1,
                'time_units': TimeUnits.SECONDS,
            }
        },
        'periodical': {
            'repeats': PeriodicalUnits.SECONDLY,
            'every': 1,
        },
        'stop': {
            'never': False,
            'after_num_repeats': 2
        }
    })
    async def some_coroutine():
        print('PING')

    await some_coroutine()  # will be printed twice


async def test_scheduler_wrong_struct():

    @scheduler({
        'stop': []
    })
    async def some_coroutine():
        return 'PING'

    with pytest.raises(SchemaInvalid):
        await some_coroutine()


async def test_concurrent_scheduler(event_loop):

    @scheduler({
        'start': {
            'relative_timeshift': {
                'delay': 1,
                'time_units': TimeUnits.SECONDS,
            }
        },
        'periodical': {
            'repeats': PeriodicalUnits.SECONDLY,
            'every': 1,
        },
        'stop': {
            'never': False,
            'after_num_repeats': 2
        }
    })
    async def some_coroutine():
        print('PING')

    @scheduler({
        'start': {
            'relative_timeshift': {
                'delay': 1,
                'time_units': TimeUnits.SECONDS,
            }
        },
        'periodical': {
            'repeats': PeriodicalUnits.SECONDLY,
            'every': 1,
        },
        'stop': {
            'never': False,
            'after_num_repeats': 2
        }
    })
    async def another_coroutine():
        print('PONG')

    # concurrent execution
    await asyncio.gather(some_coroutine(), another_coroutine())



import asyncio
import math
import functools

from krolib.parser import schedule_delta, schedule_parser
from krolib.utils import just_now


def scheduler(schedule=None):
    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(*args, **kwargs):
            if schedule:
                for dt in schedule_parser(schedule):
                    now = just_now()
                    delay_seconds = math.ceil((dt - now).total_seconds())
                    await asyncio.sleep(delay_seconds)
                    asyncio.create_task(func(*args, **kwargs))
            else:
                asyncio.create_task(func(*args, **kwargs))
        return wrapped
    return wrapper
